#!/usr/bin/perl -w
######################################################
# Author: William Lam
# 01/09/2010
# http://communities.vmware.com/docs/DOC-11902
# http://www.engineering.ucsb.edu/~duonglt/vmware/
#MC###################################################

# Enhancements/modifications provided by Michael Ceccanese 2/1/11

# This modifies the original script of dryrun and hard shutdown
# primarily geared for smaller companies who want a graceful shutdown
# of the hosts and vms and to get the script to work more quickly.
# The original scripts would work supposedly with vMA 4.0 but nobody checked 4.1.
# Anyone can take this script to make it work and then change to the more
# advanced script once they have the basics working.  That is where the
# problems seemed to lie for a first time user of the script like me.  I also
# modified the sendmail subroutine just making it very basic.  I suggest you
# run the script in dryrun mode and make sure you are at least getting emails.
#
# These modifications were mainly put in place to provide a lot more
# documentation but to help make it easier to get your first ESX host which
# is running vMA to shutdown gracefully.  I hope the rest that I added helps
# when adding other ESX hosts with VMs or without VMs.
# Additionally, a Glob file is created in /tmp to help track how far you got
# while debugging just in case an email did not get sent.
# Good luck to all.
#MC###################################################
#RB###################################################

# Modifications provided by Ruediger Brink 3/3/12

# This modifies the original script to run on ESXi 5
# and Shutdown (n) ESX and ESXi hosts which are added
# with "vifp addserver <ip-number>".
# Be careful that you add at least the vMA host!
#RB###################################################
######################################################
use strict;
use warnings;
use POSIX qw(ceil floor);
use VMware::VmaTargetLib;
use Sys::Hostname;
use Net::SMTP;
use Pod::Usage;
use File::Temp qw(tempfile tempdir);
######################################################
# List of ESX(i) hosts to exclude
######################################################
my @exclude_hosts = (
);
######################################################
# EMAIL CONF
######################################################
#MC#################
my $MCLogFile = "";
####################
#RB##################################################
# Set $RBLogFile to "" if you don't want a Logfile
#RB#
my $RBLogFile = "/home/vi-admin/rb-vma.log";
my $log_output = "";
#RB##################################################
my $EMAIL_HOST = 'mgr.riello.zip';
my $EMAIL_DOMAIN = 'riello.zip';
my $EMAIL_TO = 'Receiver <receiver@riello-ups.com>';
my $EMAIL_FROM = 'Sender <sender@riello-ups.com>';
#MC#################
#my $vma_name = "vSphere Management Assistant (vMA)";
#MC#################
my %opts = (
  host_operation => {
    type => "=s",
    help => "Host Operation to perform [shutdown|standby|autoquery|dryrun]",
    required => 0,
    default => "dryrun",
    },
  vm_operation => {
    type => "=s",
    help => "VM Operation to perform [suspend|shutdown|auto]",
    required => 0,
    default => "suspend",
    },
  vma_name => {
    type => "=s",
    help => "Display name of the vMA VM running this shutdown script",
    required => 0,
    default => "vSphere Management Assistant (vMA)",
    },
  timeout => {
    type => "=s",
    help => "Timeout value before shutting down all hosts (millisecond)",
    required => 0,
    default => 15,
    },
  sendmail => {
    type => "=s",
    help => "Send shutdown log as mail",
    required => 0,
    default => "no",
    },
  ups_host => {
    type => "=s",
    help => "IP number of vMA host",
    required => 1,
    },
  mail_timeout => {
    type => "=s",
    help => "Time to let sendmail send the mail",
    required => 0,
    default => 30,
    },
  host_repeat => {
    type => "=s",
    help => "How often the host_timeout is repeated before is canceled",
    required => 0,
    default => 5,
    },
);

# validate options, and connect to the server
Opts::add_options(%opts);
Opts::parse();
Opts::set_option("passthroughauth", 1);
Opts::validate();

my $host_operation = Opts::get_option('host_operation');
my $vm_operation = Opts::get_option('vm_operation');
my $vma_name = Opts::get_option('vma_name');
my $timeout = Opts::get_option('timeout');
my $sendmail = Opts::get_option('sendmail');
my $mail_timeout = Opts::get_option('mail_timeout');
my $host_repeat = Opts::get_option('host_repeat');
chomp($vma_name);
#RB############################################
my $ups_host = Opts::get_option('ups_host');
my $totalHosts = 0;
#RB############################################
my $vma_host_name = "";

######################################################
######## PLEASE DO NOT MODIFY PAST THIS LINE #########
######################################################
# log level
my %log_level=(
  "debug"   => 1,
  "info"  => 2,
  "warn"  => 3,
  "error"   => 4,
  "fatal"   => 5,
);
#my $LOGLEVEL = $log_level{$loglevel};
my $LOGLEVEL = $log_level{"info"};
my $vma_login;
my $host_login;
my @hosts = ();
#MC###################################################
my @hosts2 = ();
#MC###################################################
my %validhosts = ();
# MC moved numOfVMsOn initializing here so I wouldn't have a print error
my $numOfVMsOn = 0;
# MC
# MC##########################################
# This has been added to make sure that the host can log on to the vma at the end
my $host_HASH=();
# This has been added to figure out if the only VM running is vMA
# If so, then do not count it as a powered on VM because it is
# shut down later in the code
my $vma_name2;
# MC##########################################
my $recommended_timeout = 0;
######################################################
#process and get hosts managed by vMA
######################################################
if (!$log_output) {
  if ($sendmail eq 'yes') {
    $log_output = "/tmp/".tempfile();
  } else {
    $log_output = "&STDOUT";
  }
}

@hosts = VmaTargetLib::enumerate_targets();
#MC###################################################
@hosts2 = @hosts;
#MC###################################################

summaryStart();

#MC# GetHostsAndVms goes through each host and shuts down their VMs
#  It does not shutdown the vMA
#MC#
getHostsAndVms();
#MC# allow a few  seconds for all hosts to finish shutting down their VMs
# comment the line for processing faster during initial testing
sleep 10;
#MC#

verifyAndShutdownHosts();

#summaryEnd();

######################################################
#
#          HELPER FUNCTIONS
#
######################################################

sub verifyAndShutdownHosts {
  &log("info","Starting verifyAndShutdownHosts()");
  if($host_operation ne 'autoquery') {
    &log("info","Verifying all VMs are either suspended or powered off\n");
    while($totalHosts > 0) {
      for my $hostname (keys %validhosts) {
        if($validhosts{$hostname} eq "yes") {
          eval {
            &log("debug", "Login by vi-fastpass to: \"$hostname\"");
            $host_login=VmaTargetLib::query_target($hostname);
            $host_login->login();
          };
          if($@) {
            &log("info","Error: Unable to login to host: \"$hostname\"! Ensure host is being managed by vMA!");
            $totalHosts -= 1;
          } else {
            my $host_view = Vim::find_entity_view(view_type => 'HostSystem');
            my $vms = Vim::get_views(mo_ref_array => $host_view->vm, properties => ['name','runtime.powerState']);
            my $totalVMs = 0;
            $numOfVMsOn = 0;
            foreach(@$vms) {
              $totalVMs += 1;
              if($_->{'runtime.powerState'}->val eq 'poweredOn' && $_->{'name'} ne $vma_name) {
                $numOfVMsOn += 1;
              }
            }
            # remove hosts that have no more powered on VMs so
            # they're not checked again
            if($numOfVMsOn eq 0) {
              $validhosts{$hostname} = "no";
              $totalHosts -= 1;
              &log("info","Host: \"$hostname\" has $numOfVMsOn of $totalVMs VMs powered on and is ready!\n");
            } else {
              &log("info","Host: \"$hostname\" has $numOfVMsOn of $totalVMs VMs still powered on!\n");
              Util::disconnect();
              if($host_operation eq 'dryrun') {
                $totalHosts -=1;
              }
            }
          }
        }
      }
      if($totalHosts > 0 and $host_operation ne 'dryrun') {
        # RB#################################################################
        # Gives hosts $timeout seconds time to have their VMs shut down each
        # $timeout seconds represents how many times you said this could be
        # attempted by the timeout number passed to this script
        # RB#################################################################
        &log("info","Wait for $totalHosts host(s) to Shutdown the VMs - TIMEOUT set to: $timeout sec");
        $host_repeat -= 1;
        sleep $timeout;
      }
      if($host_operation eq 'dryrun') {
        $host_repeat -= 1;
      }
      if($host_repeat eq 0) {
        $totalHosts = 0;
      }
      &log("info","totalHosts: $totalHosts  host_repeat: $host_repeat  timeout: $timeout");
    }
    &log("info","Verification stage completed!\n");
    foreach my $hostname (keys %validhosts) {
      # MC ############################################
      # do not know what this is for so I replaced it
      # shutting down all hosts except for vMA host because this script is on vMA
      #      if (!defined $vma_host_name);
      $host_login=VmaTargetLib::query_target($hostname);
      $host_login->login();
      my $host_view=Vim::find_entity_view(view_type => 'HostSystem');
      # RB####################################################
      # send mail before shutdown the vMA and host. You need
      # at least 30 sec. delay to send this mail!
      # RB####################################################
      if ($hostname eq $ups_host) {
        &log("info","finally Shutdown vMA host => \"$hostname\"");
        &log("info","Putting \"$hostname\" into \"$host_operation mode\"");
        eval {
          &summaryEnd();
        };
        if($@) {
          &log("info","ERROR: $@");
        }
      } else {
        &log("info","Putting \"$hostname\" into \"$host_operation mode\"");
      }
      # RB#
      # MC ############################################
      &shutdownOrStandbyHost($hostname,$host_view);
      Util::disconnect();
      # MC ############################################
    }
  }
}

sub shutdownOrStandbyHost {
  &log("info","Starting shutdownOrStandbyHost()");
  my ($hostname,$host_view) = @_;
  if($host_operation eq 'standby' && $host_view->capability->standbySupported eq 'true') {
    if($host_operation ne 'dryrun' && $host_operation ne 'autoquery') {
      eval {
        $host_view->PowerDownHostToStandBy_Task(timeoutSec => 300)
      };
      if($@) {
        &log("info","ERROR: $@")
      }
    } else {
      &log("info","Pretend shutdown standby of \"$hostname\"");
    }
    &log("info","\"$hostname\" is going into standby mode!");
  } else {
    if($host_operation ne 'dryrun' && $host_operation ne 'autoquery') {
      eval {
        #&log("info","$host_view->ShutdownHost(force => 1)");
        $host_view->ShutdownHost(force => 1);
      };
      if($@) {
        &log("info","ERROR: $@")
      }
    } else {
      &log("info","Pretend shutdown of \"$hostname\"");
    }
    if($hostname ne $ups_host) {
      &log("info","\"$hostname\" is now shutting down!");
    }
  }
}

sub getHostsAndVms {
  &log("info","Starting getHostsAndVMs()");
  foreach my $host (@hosts) {
    my $hostname = $host->name();
    #MC ######################
    # host_HASH is to make sure we get the last ESX hash which would be the vma ESX
    # so that we can log on to the ESX so we can shutdown the vma at the end of
    # the script in VerifyAndShutdownHosts
    $host_HASH=$host;
    &log("info","Host is \"$host\"");
    &log("info","Hostname is \"$hostname\"");
    #MC ######################
    if (! grep( /^$hostname/,@exclude_hosts ) ) {
      &log("info","Found host: \"". $hostname . "\"");
      &log("debug", "Main: Login by vi-fastpass to: " . $hostname);
      my $host_type;
      eval {
        $host->login();
        #validate ESX/ESXi host
        my $content = Vim::get_service_content();
        $host_type = $content->about->apiType;
      };
      if($@) {
        &log("info","Error: Unable to login to host: \"$hostname\"! Ensure host is being managed by vMA!\n");
      } else {
        if($host_type eq 'HostAgent') {
          $totalHosts += 1;
          $validhosts{$hostname} = "yes";
          my $host_view = Vim::find_entity_view(view_type => 'HostSystem');
          if($host_operation eq 'autoquery') {
            &autoquery($host_view);
          } else {
            my $vms=Vim::get_views(mo_ref_array => $host_view->vm);
            &log("info","Begin $vm_operation operation on VMs ...");
            if($vm_operation eq 'auto' and $hostname ne $ups_host) {
              &autoShutdownVMs($host_view);
            } else {
              &shutdownVMs($vms);
            }
            &log("info","$vm_operation operation complete!");
          }
        } else {
          &log("info","Host: \"$hostname\" is not an ESX(i) host and will be ignored");
        }
        &log("debug", "Main: Disconnect from: \"$hostname\"\n");
        Util::disconnect();
      }
    }  
  }
  &log("info","Found total host(s) of: $totalHosts");
  if($host_operation eq 'autoquery' && $vm_operation eq 'auto') {
    &log("info","RECOMMENDED_TIMEOUT_VALUE = > " . ceil(($recommended_timeout/60)) . " seconds\n");
    &summaryEnd();
  }
}

sub autoquery {
  &log("info","Starting autoquery()");
  my ($host) = @_;
  if($vm_operation eq 'auto') {
    my $autoStartMgr = Vim::get_view(mo_ref => $host->configManager->autoStartManager);
    &log("info","AUTOSTART MANAGER INFO");
    &log("info","--------------------------------------");
    &log("info","AUTOSTART ENABLED   = " . ($autoStartMgr->config->defaults->enabled ? "YES" : "NO"));
    &log("info","DEFAULT STOP ACTION = " . $autoStartMgr->config->defaults->stopAction);
    my $default_delay = $autoStartMgr->config->defaults->stopDelay;
    &log("info","DEFAULT STOP DELAY  = " . ceil(($default_delay/60)) . " minute");
    my $totalTimeInSecs = 0;
    my $powerInfo = $autoStartMgr->config->powerInfo;
    if($powerInfo) {
      foreach( sort {$a->startOrder cmp $b->startOrder} @$powerInfo) {
        my $sd = $_->stopDelay;
        my $vmname = Vim::get_view(mo_ref => $_->key, properties => ['name']);
        my $order = $_->startOrder;
        if($sd eq -1) {
          $sd = $default_delay;
        }
        &log("info","VM = " . $vmname->{'name'} . "\tORDER=" . $order . "\tSTOP DELAY=" . ceil(($sd/60)) . " minute");
        $totalTimeInSecs = $totalTimeInSecs + $sd;
      }
      &log("info","TOTAL STOP DELAY    = " . ceil(($totalTimeInSecs/60)) . " minute");
      &log("info","--------------------------------------");
    }
    $recommended_timeout = $recommended_timeout + $totalTimeInSecs;
  }
}

sub autoShutdownVMs {
  &log("info","Starting autoShutdownVMs()");
  my ($host) = @_;
  my $autoStartMgr = Vim::get_view(mo_ref => $host->configManager->autoStartManager);
  my $vms = Vim::get_views(mo_ref_array => $host->vm);
  foreach my $vm (@$vms) {
    if($vm->name eq $vma_name) {
      &log("debug","Found UPS Monitoring VM: " . $vm->name);
      my $vma_host_view = Vim::get_view(mo_ref => $vm->runtime->host);
      $vma_host_name = $vma_host_view->name;
    }
  }
  if($autoStartMgr->config->defaults->enabled) {
    &log("info","Issuing AutoStartPowerOff() ...");
    if($host_operation ne 'dryrun' && $host_operation ne 'autoquery') {
      $autoStartMgr->AutoStartPowerOff();
    } else {
      my $powerInfo = $autoStartMgr->config->powerInfo;
      foreach( sort {$a->startOrder cmp $b->startOrder} @$powerInfo) {
        my $vmname = Vim::get_view(mo_ref => $_->key, properties => ['name']);
        my $action;
        if($_->stopAction eq 'SystemDefault') {
          $action = $autoStartMgr->config->defaults->stopAction;
        } else {
          $action = $_->stopAction;
        }
        &log("info","Performing $action operation on " . $vmname->{'name'});
      }
    }
  } else {
    &log("info","Auto start/stop manager is not enabled, VMs will not be powering down!");
  }
}

sub shutdownVMs {
  &log("info","Starting shutdownVMs()");
  &log("info","vma_name is \"$vma_name\"");
  my ($vms) = @_;
  foreach my $vm (@$vms) {
    my $vmname = $vm->name;
    #MC ######################
    if($vmname eq $vma_name) {
      &log("info","Not shutting down \"$vm\" yet. This is done last.");
    } else {
      &log("info","vmname is \"$vmname\"");
      &log("info","vm is $vm");
    }
    #MC ######################
    if($vm_operation ne 'dryrun') {
      if($vmname ne $vma_name) {
        if($vm->runtime->powerState->val eq 'poweredOn') {
          if($vm_operation eq 'suspend') {
            eval {
              &log("info","Suspending \"$vmname\" ...");
              if($host_operation ne 'dryrun' && $host_operation ne 'autoquery') {
                $vm->SuspendVM();
              }
            };
            if($@) {
              &log("info","Unable to suspend \"$vmname!\"");
            }
          } else {
            &log("info","Status of VMTools = " . $vm->guest->toolsStatus->val);
            if(defined($vm->guest) && ($vm->guest->toolsStatus->val eq 'toolsOld' || $vm->guest->toolsStatus->val eq 'toolsOk') ) {
              eval {
                if($host_operation ne 'dryrun' && $host_operation ne 'autoquery') {
                  &log("info","Shutdown \"$vmname\" via VMware Tools...");
                  $vm->ShutdownGuest();
                } else {
                  &log("info","Simulate shutdown \"$vmname\" via VMware Tools...");
                }
              };
              if($@) {
                &log("info","Unable to shutdown \"$vmname!\"");
              }
            } else {
              eval {
                if($host_operation ne 'dryrun' && $host_operation ne 'autoquery') {
                  &log("info","Hard powering off \"$vmname\", no VMware Tools found ...");
                  $vm->PowerOffVM();
                } else {
                  &log("info","Simulate hard powering off \"$vmname\", no VMware Tools found ...");
                }
              };
              if($@) {
                &log("info","Unable to hard power off \"$vmname!\"");
              }
            }
          }
        } else {
          &log("info","\"$vmname\" was already powered off!");
        }
      } else {
        #MC ######################
        &log("info","Found UPS Monitoring VM: \"$vmname\"");
        # Do not want to shut down vMA here
        #$vm->ShutdownGuest();
        #MC ######################
        my $vma_host_view = Vim::get_view(mo_ref => $vm->runtime->host);
        $vma_host_name = $vma_host_view->name;
        &log("info","vma_host_name in Shutdown is \"$vma_host_name\"");
      }
    }
  }
}

sub summaryStart {
  if($host_operation eq 'dryrun' || $host_operation eq 'autoquery') {
    &log("info","================= DRYRUN MODE ENABLED ==================");
  } else {
    &log("info","=========== STARTING ghettoHostShutdown5.pl ============");
  }
  #log ups_host
  &log("info","UPS_HOST......: $ups_host");
  #host selection
  &log("info","HOST_LIST.....: AUTOMATIC");
  #host operation
  &log("info","HOST_OPERATION: $host_operation");
  #host repeat
  &log("info","HOST_REPEAT...: $host_repeat");
  #host timeout
  &log("info","HOST_TIMEOUT..: $timeout");
  #vm operation
  &log("info","VM_OPERATION..: $vm_operation");
  #log vma_name
  &log("info","VMA_NAME......: $vma_name");
  #host timeout
  &log("info","SENDMAIL......: $sendmail");
  #mail timeout
  &log("info","MAIL_TIMEOUT..: $mail_timeout");

  &log("info","========================================================\n");
} 

sub summaryEnd {
  #&log("info","");
  if($host_operation eq 'dryrun' || $host_operation eq 'autoquery') {
    &log("info","================= DRYRUN MODE COMPLETED ================\n");
  } else {
    &log("info","========== COMPLETED ghettoHostShutdown5.pl ============\n");
  }
  if($sendmail eq 'yes') {
    &sendMail();
  }
}

# MC####################
# had trouble with all of the functionality of the original sendmail
# I made this one simpler
# Please review original scripts just in case they are more compatible
# with your email systems
# MC####################
sub sendMail {
  my $smtp = Net::SMTP->new($EMAIL_HOST,Hello=>$EMAIL_DOMAIN,Timeout=>30,);
  #my $smtp = Net::SMTP->new('192.168.33.7');

  $smtp->mail($ENV{USER});
  $smtp->to($EMAIL_TO);
  $smtp->data();
  $smtp->datasend("To: $EMAIL_TO\n");
  # RB################
  #$smtp->datasend("From: ghettoHostShutdown5.pl\n");
  $smtp->datasend("From: $EMAIL_FROM\n");
  # RB#
  $smtp->datasend("Subject: UPS \"$vm_operation\" VMs and \"$host_operation\" host(s)\n");
  # This give it the attachment name
  #$smtp->datasend("Content-Type:application/text;name=UPSShuttingDownHost.txt\n");
  # MC###############
  #$smtp->datasend($log_output);
  $smtp->datasend($MCLogFile);
  # MC#
  $smtp->datasend("\n");
  $smtp->dataend();
  $smtp->quit;
  # RB##########
  sleep $mail_timeout; # give sendmail time to make his work
}

sub timeStamp {
  my ($date_format) = @_;
  my %dttime = ();
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
  my $my_time;
  my $time_string;
  ### begin_: initialize DateTime number formats
  $dttime{year } = sprintf "%04d",($year + 1900); ## 4 digits to specify year
  $dttime{mon  } = sprintf "%02d",($mon + 1);   ## 0pad months
  $dttime{mday } = sprintf "%02d",$mday;      ## 0pad day of the month
  $dttime{wday } = sprintf "%02d",$wday + 1;    ## 0pad day of week; sunday=1;
  $dttime{yday } = sprintf "%02d",$yday;      ## 0pad nth day of year
  $dttime{hour } = sprintf "%02d",$hour;      ## 0pad hour
  $dttime{min  } = sprintf "%02d",$min;       ## 0pad minutes
  $dttime{sec  } = sprintf "%02d",$sec;       ## 0pad seconds
  $dttime{isdst} = $isdst;
  if($date_format eq 'MDYHMS') {
    $my_time = "$dttime{mon}-$dttime{mday}-$dttime{year} $dttime{hour}:$dttime{min}:$dttime{sec}";
    $time_string = $my_time." -- ";
  } elsif ($date_format eq 'YMD') {
    $my_time = "$dttime{year}-$dttime{mon}-$dttime{mday}";
    $time_string = $my_time;
  }
  return $time_string;
}

# Print message to log output with time stamp if appropriate to log level
sub log {
  my($logLevel, $message) = @_;
  if($message ne ' ') {
    #MC####################
    $MCLogFile = "$MCLogFile\n" . timeStamp('MDYHMS') . "$message";
    #######################
    if($RBLogFile) {
      system("echo '" . timeStamp('MDYHMS') . "$message' >> $RBLogFile");
    }
    open(MYLOG,">>$log_output") || die("Cannot open $log_output: $?");
    if ($LOGLEVEL <= $log_level{$logLevel}) {
      print MYLOG timeStamp('MDYHMS'), $logLevel, ": ", $message, "\n";
    }
    close(MYLOG);
  }
}
